# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module extends the permittivity module from the chemicals's library:
# https://github.com/CalebBell/chemicals
# Copyright (C) 2020 Caleb Bell <Caleb.Andrew.Bell@gmail.com>
#
# This module is under a dual license:
# 1. The UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
# 
# 2. The MIT open-source license. See
# https://github.com/CalebBell/chemicals/blob/master/LICENSE.txt for details.
import numpy as np
from ..utils import forward
from chemicals import volume as vol
from chemicals.virial import (
    BVirial_Pitzer_Curl, 
    BVirial_Abbott, 
    BVirial_Tsonopoulos, 
    BVirial_Tsonopoulos_extended
)
from chemicals.dippr import EQ105
# from .electrochem import _Laliberte_Density_ParametersDict, Laliberte_Density
from ..base import (functor, InterpolatedTDependentModel, 
                    TPDependentModel, TPDependentHandleBuilder,
                    PhaseTPHandleBuilder)
from .data import (
    VDI_saturation_dict,
    lookup_VDI_tabular_data,    
    rho_data_COSTALD, 
    rho_data_SNM0,
    rho_data_Perry_8E_105_l, 
    rho_data_VDI_PPDS_2, 
    rho_data_CRC_inorg_l,
    rho_data_CRC_inorg_l_const,
    rho_data_CRC_inorg_s_const, 
    rho_data_CRC_virial
)

vol.__all__.extend([
    'volume_handle',
    'volume_solid_handle',
    'volume_liquid_handle',
    'volume_gas_handle',
])



# %% Liquids

Yen_Woods_saturation = functor(vol.Yen_Woods_saturation, 'V.l')
Rackett = functor(vol.Rackett, 'V.l')
Yamada_Gunn = functor(vol.Yamada_Gunn, 'V.l')
Townsend_Hales = functor(vol.Townsend_Hales, 'V.l')
Bhirud_normal = functor(vol.Bhirud_normal, 'V.l')
COSTALD = functor(vol.COSTALD, 'V.l')
Campbell_Thodos = functor(vol.Campbell_Thodos, 'V.l')
SNM0 = functor(vol.SNM0, 'V.l')
CRC_inorganic = functor(vol.CRC_inorganic, 'V.l')
volume_VDI_PPDS = functor(vol.volume_VDI_PPDS , 'V.l')
COSTALD_compressed = functor(vol.COSTALD_compressed , 'V.l')

def EQ105_hook(self, T, kwargs):
    if kwargs.get('molar_volume'):
        del kwargs['molar_volume']
        V_inv = self.function(T, **kwargs)
        value = 1. / V_inv
        kwargs['molar_volume'] = True
    else:
        value = self.function(T, **kwargs)
    return value
EQ105.functor.hook = EQ105_hook

def COSTALD_compressed_hook(self, T, P, kwargs):
    kwargs = kwargs.copy()
    Psat = kwargs['Psat']
    Vs = kwargs['Vs']
    if callable(Psat): kwargs['Psat'] = Psat(T)
    if callable(Vs): kwargs['Vs'] = Vs(T, P)
    return self.function(T, P, **kwargs)
COSTALD_compressed.functor.hook = COSTALD_compressed_hook

@TPDependentHandleBuilder('V.l')
def volume_liquid_handle(handle, CAS, MW, Tb, Tc, Pc, Vc, Zc,
                         omega, Psat, eos, dipole, has_hydroxyl):
    all_ = all
    add_model = handle.add_model
    if CAS in rho_data_VDI_PPDS_2:
        MW, Tc_, rhoc, A, B, C, D = rho_data_VDI_PPDS_2[CAS]
        data = (Tc, rhoc, A, B, C, D, MW)
        add_model(volume_VDI_PPDS.functor.from_args(data), 0., Tc_)
    if CAS in rho_data_Perry_8E_105_l:
        C1, C2, C3, C4, Tmin, Tmax = rho_data_Perry_8E_105_l[CAS]
        data = (C1, C2, C3, C4, 0)
        f = EQ105.functor.from_args(data)
        f.molar_volume = True
        add_model(f, Tmin, Tmax)
    if CAS in VDI_saturation_dict:
        Ts, Vls = lookup_VDI_tabular_data(CAS, 'Volume (l)')
        model = InterpolatedTDependentModel(Ts, Vls,
                                            Tmin=Ts[0], Tmax=Ts[-1])
        add_model(model)
    data = (Tb, Tc, Pc, MW, dipole, has_hydroxyl)
    if all_([i is not None for i in data]):
        add_model(Campbell_Thodos.functor.from_args(data), 0, Tc)
    if CAS in rho_data_CRC_inorg_l:
        MW, rho, k, Tm, Tmax = rho_data_CRC_inorg_l[CAS]
        data = (MW, rho, k, Tm)
        add_model(CRC_inorganic.functor.from_args(data), Tm, Tmax)
    data = (Tc, Vc, Zc)
    if all_(data):
        add_model(Yen_Woods_saturation.functor.from_args(data), 0, Tc)
    data = (Tc, Pc, Zc)
    if all_(data):
        add_model(Rackett.functor.from_args(data), 0, Tc)
    data = (Tc, Pc, omega)
    if all_(data):
        add_model(Yamada_Gunn.functor.from_args(data), 0, Tc)
        add_model(Bhirud_normal.functor.from_args(data), 0, Tc)
    data = (Tc, Vc, omega)
    if all_(data):
        add_model(Townsend_Hales.functor.from_args(data), 0, Tc)
        if CAS in rho_data_SNM0:
            SNM0_delta_SRK = float(rho_data_SNM0.get(CAS, 'delta_SRK'))
            data = (Tc, Vc, omega, SNM0_delta_SRK)
            add_model(SNM0.functor.from_args(data), 0, Tc)
    if CAS in rho_data_CRC_inorg_l_const:
        Vl = float(rho_data_CRC_inorg_l_const.get(CAS, 'Vm'))
        add_model(Vl, 0., Tc, name="CRC inorganic liquid constant")
    if Tc and Pc and CAS in rho_data_COSTALD:
        Zc_ = rho_data_COSTALD.get(CAS, 'Z_RA')
        if not np.isnan(Zc_): Zc_ = float(Zc_)
        data = (Tc, Pc, Zc_)
        add_model(Rackett.functor.from_args(data), 0., Tc)
        # Roughly data at STP; not guaranteed however; not used for Trange
    data = (Tc, Vc, omega)
    if all_(data) and CAS in rho_data_COSTALD:
        add_model(COSTALD.functor.from_args(data), 0, Tc)
    data = (Tc, Pc, omega)
    if all_(data):
        data = (Psat, Tc, Pc, omega, handle)
        add_model(COSTALD_compressed.functor.from_args(data), 50, 500)
vol.volume_liquid_handle = volume_liquid_handle

# %% Gases

ideal_gas = vol.ideal_gas

ideal_gas_model = TPDependentModel(vol.ideal_gas,
                                   0, 10e6,
                                   0, 10e12,
                                   var='V.g')

@forward(vol)
@functor(var='V.g')
def Tsonopoulos_extended(T, P, Tc, Pc, omega, a=0, b=0,
                         species_type='', dipole=0, order=0):
    return ideal_gas(T, P) + BVirial_Tsonopoulos_extended(T, Tc, Pc, omega, a, b,
                                                          species_type, dipole, order)

@forward(vol)
@functor(var='V.g') 
def Tsonopoulos(T, P, Tc, Pc, omega):
    return ideal_gas(T, P) + BVirial_Tsonopoulos(T, Tc, Pc, omega)

@forward(vol)
@functor(var='V.g')
def Abbott(T, P, Tc, Pc, omega):
    return ideal_gas(T, P) + BVirial_Abbott(T, Tc, Pc, omega)

@forward(vol)
@functor(var='V.g')
def Pitzer_Curl(T, P, Tc, Pc, omega):
    return ideal_gas(T, P) + BVirial_Pitzer_Curl(T, Tc, Pc, omega)
    
@forward(vol)
@functor(var='V.g')
def CRCVirial(T, P, a1, a2, a3, a4, a5):
    t = 298.15/T - 1.
    return ideal_gas(T, P) + (a1 + t*(a2 + t*(a3 + t*(a4 + a5*t))))/1e6

@TPDependentHandleBuilder('V.g')
def volume_gas_handle(handle, CAS, Tc, Pc, omega, eos):
    add_model = handle.add_model
    # no point in getting Tmin, Tmax
    if all((Tc, Pc, omega)):
        data = (Tc, Pc, omega)
        add_model(Tsonopoulos_extended.functor.from_args(data))
        add_model(Tsonopoulos.functor.from_args(data))
        add_model(Abbott.functor.from_args(data))
        add_model(Pitzer_Curl.functor.from_args(data))
    if CAS in rho_data_CRC_virial:
        add_model(CRCVirial.functor.from_args(rho_data_CRC_virial[CAS]))
    add_model(ideal_gas_model)
vol.volume_gas_handle = volume_gas_handle

# %% Solids

Goodman = functor(vol.Goodman, 'V.s')

# TODO: Make a data table for new additions (not just for common sugars).
# Source: PubChem
sugar_solid_densities = {
    '50-99-7': 0.0001125975, 
    '57-48-7': 0.0001063495,
    '3458-28-4': 0.00011698,
    '25990-60-7': 9.84459e-05,
    '59-23-4': 0.000120104,   
}

@TPDependentHandleBuilder('V.s')
def volume_solid_handle(handle, CAS):
    if CAS in rho_data_CRC_inorg_s_const:
        CRC_INORG_S_Vm = float(rho_data_CRC_inorg_s_const.get(CAS, 'Vm'))
        handle.add_model(CRC_INORG_S_Vm, 0, 1e6, 0, 1e12)
    elif CAS in sugar_solid_densities: 
        handle.add_model(sugar_solid_densities[CAS], 0, 1e6, 0, 1e12)
vol.volume_solid_handle = volume_solid_handle

vol.volume_handle = PhaseTPHandleBuilder('V',
                                     volume_solid_handle, 
                                     volume_liquid_handle,
                                     volume_gas_handle)




# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2021, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
from collections import deque
from numpy import inf as infinity
from .thermo_model import (ThermoModel,
                           TDependentModel,
                           TPDependentModel,
                           thermo_model,
                           create_axis_labels)
from ..exceptions import DomainError
from ..units_of_measure import definitions
from .functor import functor_lookalike
import matplotlib.pyplot as plt

__all__ = ('ThermoModelHandle',
           'TDependentModelHandle',
           'TPDependentModelHandle')

# %% Utilities

def find_model_by_name(models, name):
    for model in models:
        if model.name == name: return model
    raise LookupError(f'no model with name {name}')
    
def find_model_index_by_name(models, name):
    for index, model in enumerate(models):
        if model.name == name: return index
    raise LookupError(f'no model with name {name}')
    
def no_valid_model(chemical, var):
    definition = definitions.get(var)
    msg = f"{chemical} (CAS: {chemical.CAS}) has no valid " if chemical else "no valid "
    msg += f"{definition.lower()} model" if definition else "model"
    return msg

def as_model_index(models, key):
    isa = isinstance
    if isa(key, int):
        index = key
    elif isa(key, str):
        name = key
        index = find_model_index_by_name(models, name)
    elif isa(key, ThermoModel):
        index = models.index(key)
    else:
        raise TypeError(
            'key must be either the index [int], name [str], '
            'or the model itself [ThermoModel]')
    return index

def as_model(models, key):
    """Return model given the index, name, or the model itself."""
    isa = isinstance
    if isa(key, int):
        index = key
        model = models[index]
    elif isa(key, str):
        name = key
        model = find_model_by_name(models, name)
    elif isa(key, ThermoModel):
        model = key
        assert model in models, "''"
    else:
        raise TypeError(
            'key must be either the index [int], name [str], '
            'or the model itself [ThermoModel]')
    return model


# %% Handles

@functor_lookalike
class ThermoModelHandle:
    __slots__ = ('_chemical', '_var', '_models')
    
    @property
    def chemical(self):
        """[Chemical] Parent chemical."""
        return self._chemical
    @property
    def var(self):
        """[str] Return variable."""
        return self._var
    @property
    def models(self):
        """Deque[ThermoModel] All models."""
        return self._models
    
    def set_value(self, var, value):
        for model in self._models: model.set_value(var, value)
    
    def plot_vs_T(self, T_range=None, T_units=None, units=None, 
                  P=101325, label_axis=True, **plot_kwargs): # pragma: no cover
        Ts, Ys = self.tabulate_vs_T(T_range, T_units, units, P)
        plt.plot(Ts, Ys, **plot_kwargs)
        if label_axis: create_axis_labels('T', T_units, self._var, units)
        plt.legend()
        
    def plot_vs_P(self, P_range=None, P_units=None, units=None,
                  T=298.15, label_axis=True, **plot_kwargs): # pragma: no cover
        Ps, Ys = self.tabulate_vs_P(P_range, P_units, units, T)
        plt.plot(Ps, Ys, **plot_kwargs)
        if label_axis: create_axis_labels('P', P_units, self._var, units)
        plt.legend()
    
    def plot_models_vs_T(self, T_range=None, T_units=None, units=None,
                         P=101325, label_axis=True, **plot_kwargs): # pragma: no cover
        for model in self: model.plot_vs_T(T_range, T_units, units, P, False, **plot_kwargs)
        if label_axis: create_axis_labels('T', T_units, self._var, units)
        plt.legend()
    
    def plot_models_vs_P(self, P_range=None, P_units=None, units=None,
                         T=298.15, label_axis=True, **plot_kwargs): # pragma: no cover
        for model in self: model.plot_vs_P(P_range, P_units, units, T, False, **plot_kwargs)
        if label_axis: create_axis_labels('P', P_units, self._var, units)
        plt.legend()
    
    def __init__(self, var, models=None):
        self._chemical = None
        self._var = var
        self._models = deque(models) if models else deque()
    
    def __getitem__(self, index):
        models = self._models
        return models[as_model_index(models, index)]
    
    def __setitem__(self, index, model):
        if not isinstance(model, ThermoModel):
            raise TypeError("a 'ThermoModelHandle' object may only "
                            "contain 'ThermoModel' objects")
        models = self._models
        models[as_model_index(models, index)] = model
	
    def __iter__(self):
        return iter(self._models)
    
    def __bool__(self):
        return bool(self._models)
    
    def copy(self):
        return self.__class__(self._var, self._models.copy())
    __copy__ = copy
    
    def set_model_priority(self, key, priority=0):
        """
        Set model priority.

        Parameters
        ----------
        key : int, str, or ThermoModel
            Index, name, or the model itself.
        priority : int, optional
            Index to place model. The default is 0.

        """
        models = self._models
        model = as_model(models, key)
        models.remove(model)
        models.insert(priority, model)
    
    def move_up_model_priority(self, key, priority=0):
        """
        Move up model priority.
        
        Parameters
        ----------
        key : int, str, or ThermoModel
            Index, name, or the model itself.
        priority : int, optional
            Index to place model. The default is 0.
            
        """
        index = as_model_index(self._models, key)
        self._models.rotate(priority - index)
    
    def add_model(self, evaluate=None,
                  Tmin=None, Tmax=None,
                  Pmin=None, Pmax=None,
                  name=None, top_priority=False, 
                  **kwargs):
        """
        Add model to handle.

        Parameters
        ----------
        evaluate : ThermoModel, function, or float
            Model function or constant value.
        Tmin : float, optional
            Minimum temperature the that model can be used [K]. Defaults to 0.
        Tmax : float, optional
            Maximum temperature the that model can be used [K]. Defaults to infinity.
        Pmin : float, optional
            Minimum pressure the that model can be used [Pa]. Defaults to 0.
        Pmax : float, optional
            Maximum pressure the that model can be used [Pa]. Defaults to infinity.
        name : str, optional
            Name of the model. Defaults to function name.
        top_priority : int, optional
            Whether to add the model in 0th index/priority. The default is False.
        
        Other Parameters
        ----------------
        integrate_by_T : function(Ta, Tb, ...)
            Analytical integration function (to replace numerical integration).
        integrate_by_T_over_T : function(Ta, Tb, ...)
            Analytical integration function (to replace numerical integration).
        
        Notes
        -----
        This method can be used as a function decorator, but the function is
        not replaced with the model.

        """
        if evaluate is None:
            return lambda evaluate: self.add_model(evaluate,
                                                   Tmin, Tmax,
                                                   Pmin, Pmax,
                                                   name, top_priority,
                                                   **kwargs)
        if isinstance(evaluate, ThermoModel):
            model = evaluate
        elif isinstance(evaluate, ThermoModelHandle):
            raise TypeError("ThermoModelHandle object cannot be added as a "
                            "model; only ThermoModel objects, functions, "
                            "and constants can be added")
        else:
            model = thermo_model(evaluate, Tmin, Tmax, Pmin, Pmax,
                                 name, self._var, **kwargs)
        if top_priority:
            self._models.appendleft(model)
        else:
            self._models.append(model)    
        
        if callable(evaluate): return evaluate
       
    def remove(self, key):
        """
        Remove model
        
        Parameters
        ----------
        key : int, str, or ThermoModel
            Index, name, or the model itself.
            
        """
        model = as_model(self._models, key)
        self._models.remove(model)
       
    def show(self):
        info = f"{self}\n"
        if self._models:
            models = "\n".join([f'[{i}] {model.name}'
                                for i, model in enumerate(self._models)])
        else:
            models = "(no models available)"
        print(info + models)
        
    _ipython_display_ = show

    
class TDependentModelHandle(ThermoModelHandle):
    __slots__ = ()
    Pmin = 0
    Pmax = infinity
    tabulate_vs_T = TDependentModel.tabulate_vs_T
    tabulate_vs_P = TDependentModel.tabulate_vs_P
        
    @property
    def Tmin(self):
        return min([i.Tmin for i in self._models])
    @property
    def Tmax(self):
        return max([i.Tmax for i in self._models])
    
    def __call__(self, T, P=None):
        for model in self._models:
            if model.indomain(T): return model.evaluate(T)
        raise DomainError(f"{no_valid_model(self._chemical, self._var)} "
                          f"at T={T:.2f} K", chemical=self._chemical)
    
    at_T = __call__
    
    def try_out(self, T, P=None):
        for model in self._models:
            if model.indomain(T):
                try: return model.evaluate(T)
                except: pass
    
    def differentiate_by_T(self, T, P=None, dT=1e-12):
        for model in self._models:
            if model.indomain(T): return model.differentiate_by_T(T, dT=dT)
        raise DomainError(f"{no_valid_model(self._chemical, self._var)} "
                         f"at T={T:.2f} K", chemical=self._chemical)
        
    def differentiate_by_P(self, T, P=None, dP=1e-12):
        return 0
        
    def integrate_by_T(self, Ta, Tb, P=None):
        integral = 0.
        defined = hasattr
        for model in self._models:
            if not defined(model, 'integrate_by_T'): continue
            Tmax = model.Tmax
            Tmin = model.Tmin
            lb_satisfied = Ta > Tmin
            ub_satisfied = Tb < Tmax
            if lb_satisfied:
                if ub_satisfied:
                    return integral + model.integrate_by_T(Ta, Tb)
                elif Ta < Tmax:
                    integral += model.integrate_by_T(Ta, Tmax)
                    Ta = Tmax
            elif ub_satisfied and Tb > Tmin:
                integral += model.integrate_by_T(Tmin, Tb)
                Tb = Tmin
        raise DomainError(f"{no_valid_model(self._chemical, self._var)} "
                         f"between T={Ta:.2f} to {Tb:.2f} K", chemical=self._chemical)
    
    def integrate_by_P(self, Pa, Pb, T):
        return (Pb - Pa) * self(T)
    
    def integrate_by_T_over_T(self, Ta, Tb, P=None):
        integral = 0.
        defined = hasattr
        for model in self._models:
            if not defined(model, 'integrate_by_T_over_T'): continue
            Tmax = model.Tmax
            Tmin = model.Tmin
            lb_satisfied = Ta > Tmin
            ub_satisfied = Tb < Tmax
            if lb_satisfied:
                if ub_satisfied:
                    return integral + model.integrate_by_T_over_T(Ta, Tb)
                elif Ta < Tmax:
                    integral += model.integrate_by_T_over_T(Ta, Tmax)
                    Ta = Tmax
            elif ub_satisfied and Tmin < Tb:
                integral += model.integrate_by_T_over_T(Tmin, Tb)
                Tb = Tmin
        raise DomainError(f"{no_valid_model(self._chemical, self._var)} "
                         f"between T={Ta:.2f} to {Tb:.2f} K", chemical=self._chemical)
    
    
class TPDependentModelHandle(ThermoModelHandle):
    __slots__ = ('_integrate_by_T_cache', '_integrate_by_T_over_T_cache')
    
    Tmin = TDependentModelHandle.Tmin
    Tmax = TDependentModelHandle.Tmax
    
    tabulate_vs_T = TPDependentModel.tabulate_vs_T
    tabulate_vs_P = TPDependentModel.tabulate_vs_P
    
    @property
    def Pmin(self):
        return min([i.Pmin for i in self._models])
    @property
    def Pmax(self):
        return max([i.Pmax for i in self._models])
    
    def __call__(self, T, P):
        for model in self._models:
            if model.indomain(T, P): return model.evaluate(T, P)
        raise DomainError(f"{no_valid_model(self._chemical, self._var)} "
                          f"at T={T:.2f} K and P={P:.0f} Pa", chemical=self._chemical)

    def at_T(self, T):
        isa = isinstance
        for model in self._models:
            if isa(model, TDependentModel) and model.indomain(T):
                return model.evaluate(T)
        raise DomainError(f"{no_valid_model(self._chemical, self._var)} "
                          f"at T={T:.2f} K", chemical=self._chemical)

    def try_out(self, T, P):
        for model in self._models:
            if model.indomain(T, P): return model.evaluate(T, P)

    def differentiate_by_T(self, T, P):
        for model in self._models:
            if model.indomain(T, P): return model.differentiate_by_T(T, P)
        raise DomainError(f"{no_valid_model(self._chemical, self._var)} "
                          f"at T={T:.2f} K and P={P:.0f} Pa", chemical=self._chemical)
            
    def differentiate_by_P(self, T, P):
        for model in self._models:
             if model.indomain(T, P): return model.differentiate_by_P(T, P)
        raise DomainError(f"{no_valid_model(self._chemical, self._var)} "
                          f"at T={T:.2f} K and P={P:.0f} Pa", chemical=self._chemical)
        
    def integrate_by_T(self, Ta, Tb, P):
        integral = 0
        defined = hasattr
        for model in self._models:
            if not (defined(model, 'integrate_by_T') and model.Pmin < P < model.Pmax): continue
            Tmax = model.Tmax
            Tmin = model.Tmin    
            lb_satisfied = Ta > Tmin
            ub_satisfied = Tb < Tmax
            if lb_satisfied:
                if ub_satisfied:
                    return integral + model.integrate_by_T(Ta, Tb, P)
                elif Ta < Tmax:
                    integral += model.integrate_by_T(Ta, Tmax, P)
                    Ta = Tmax
            elif ub_satisfied and Tmin < Tb:
                integral += model.integrate_by_T(Tmin, Tb, P)
                Tb = Tmin
        raise DomainError(f"{no_valid_model(self._chemical, self._var)} "
                          f"between T={Ta:.2f} to {Tb:.2f} K at P={P:.0f} Pa",
                          chemical=self._chemical)
    
    def integrate_by_P(self, Pa, Pb, T):
        integral = 0
        defined = hasattr
        for model in self._models:
            if not (defined(model, 'integrate_by_P')
                    and model.Tmin < T < model.Tmax): continue
            Pmin = model.Pmin
            Pmax = model.Pmax
            lb_satisfied = Pa > Pmin
            ub_satisfied = Pb < Pmax
            if lb_satisfied:
                if ub_satisfied:
                    return integral + model.integrate_by_P(Pa, Pb, T)
                elif Pa < Pmax:
                    integral += model.integrate_by_P(Pa, Pmax, T)
                    Pa = Pmax
            elif ub_satisfied and Pmin < Pb:
                integral += model.integrate_by_P(Pmin, Pb, T)
                Pb = Pmin
        raise DomainError(f"{no_valid_model(self._chemical, self._var)} "
                          f"between P={Pa:5g} to {Pb:5g} Pa ast T={T:.2f}",
                          chemical=self._chemical)
    
    def integrate_by_T_over_T(self, Ta, Tb, P):
        integral = 0
        defined = hasattr
        for model in self._models:
            if not (defined(model, 'integrate_by_T_over_T')
                    and model.Pmin < P < model.Pmax): continue
            Tmax = model.Tmax
            Tmin = model.Tmin    
            lb_satisfied = Ta > Tmin
            ub_satisfied = Tb < Tmax
            if lb_satisfied:
                if ub_satisfied:
                    return integral + model.integrate_by_T_over_T(Ta, Tb, P)
                elif Ta < Tmax:
                    integral += model.integrate_by_T_over_T(Ta, Tmax, P)
                    Ta = Tmax
            elif ub_satisfied and Tmin < Tb:
                integral += model.integrate_by_T_over_T(Tmin, Tb, P)
                Tb = Tmin
        raise DomainError(f"{no_valid_model(self._chemical, self._var)} "
                          f"between T={Ta:.2f} to {Tb:.2f} K",
                          chemical=self._chemical)
        
            
    


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2021, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
Free energy functors for Chemical objects.

"""
from .constants import R
from .base import functor, PhaseTFunctorBuilder, PhaseTPFunctorBuilder
from math import log
    
@functor(var='H')
def Enthalpy(T, Cn, T_ref, H_ref):
    return H_ref + Cn.integrate_by_T(T_ref, T)

@functor(var='S')
def Entropy(T, Cn, T_ref, S0):
    return S0 + Cn.integrate_by_T_over_T(T_ref, T)

@functor(var='H.l')
def Liquid_Enthalpy_Ref_Liquid(T, Cn_l, T_ref, H_ref):
    """Enthapy (kJ/kmol) disregarding pressure and assuming the specified phase."""
    return H_ref + Cn_l.integrate_by_T(T_ref, T)

@functor(var='H.l')
def Liquid_Enthalpy_Ref_Gas(T, Cn_l, H_int_Tb_to_T_ref_g, Hvap_Tb, Tb, H_ref):
    return H_ref - H_int_Tb_to_T_ref_g - Hvap_Tb + Cn_l.integrate_by_T(Tb, T)
    
@functor(var='H.l')
def Liquid_Enthalpy_Ref_Solid(T, Cn_l, H_int_T_ref_to_Tm_s, Hfus, Tm, H_ref):
    return H_ref + H_int_T_ref_to_Tm_s + Hfus + Cn_l.integrate_by_T(Tm, T)
    
@functor(var='H.s')
def Solid_Enthalpy_Ref_Solid(T, Cn_s, T_ref, H_ref):
    return H_ref + Cn_s.integrate_by_T(T_ref, T)

@functor(var='H.s')
def Solid_Enthalpy_Ref_Liquid(T, Cn_s, H_int_Tm_to_T_ref_l, Hfus, Tm, H_ref):
    return H_ref - H_int_Tm_to_T_ref_l - Hfus + Cn_s.integrate_by_T(Tm, T)

@functor(var='H.s')
def Solid_Enthalpy_Ref_Gas(T, Cn_s, H_int_Tb_to_T_ref_g, Hvap_Tb,
                           H_int_Tm_to_Tb_l, Hfus, Tm, H_ref):
    return H_ref - H_int_Tb_to_T_ref_g - Hvap_Tb - H_int_Tm_to_Tb_l - Hfus + Cn_s.integrate_by_T(Tm, T)
    
@functor(var='H.g')
def Gas_Enthalpy_Ref_Gas(T, Cn_g, T_ref, H_ref):
    return H_ref + Cn_g.integrate_by_T(T_ref, T)

@functor(var='H.g')
def Gas_Enthalpy_Ref_Liquid(T, Cn_g, H_int_T_ref_to_Tb_l, Hvap_Tb, 
                            Tb, H_ref):
    return H_ref + H_int_T_ref_to_Tb_l + Hvap_Tb + Cn_g.integrate_by_T(Tb, T)

@functor(var='H.g')
def Gas_Enthalpy_Ref_Solid(T, Cn_g, H_int_T_ref_to_Tm_s, Hfus,
                           H_int_Tm_to_Tb_l, Hvap_Tb, Tb, H_ref):
    return H_ref + H_int_T_ref_to_Tm_s + Hfus + H_int_Tm_to_Tb_l + Hvap_Tb + Cn_g.integrate_by_T(Tb, T)


EnthalpyRefLiquid = PhaseTFunctorBuilder('H',
                                        Solid_Enthalpy_Ref_Liquid.functor,
                                        Liquid_Enthalpy_Ref_Liquid.functor,
                                        Gas_Enthalpy_Ref_Liquid.functor)

EnthalpyRefSolid = PhaseTFunctorBuilder('H',
                                       Solid_Enthalpy_Ref_Solid.functor,
                                       Liquid_Enthalpy_Ref_Solid.functor,
                                       Gas_Enthalpy_Ref_Solid.functor)

EnthalpyRefGas = PhaseTFunctorBuilder('H',
                                     Solid_Enthalpy_Ref_Gas.functor,
                                     Liquid_Enthalpy_Ref_Gas.functor,
                                     Gas_Enthalpy_Ref_Gas.functor)

@functor(var='S.l')
def Liquid_Entropy_Ref_Liquid(T, Cn_l, T_ref, S0):
    """Enthapy (kJ/kmol) disregarding pressure and assuming the specified phase."""
    return S0 + Cn_l.integrate_by_T_over_T(T_ref, T)

@functor(var='S.l')
def Liquid_Entropy_Ref_Gas(T, Cn_l, S_int_Tb_to_T_ref_g, Svap_Tb, Tb, S0):
    return S0 - S_int_Tb_to_T_ref_g - Svap_Tb + Cn_l.integrate_by_T_over_T(Tb, T)
    
@functor(var='S.l')
def Liquid_Entropy_Ref_Solid(T, Cn_l, S_int_T_ref_to_Tm_s, Sfus, Tm, S0):
    return S0 + S_int_T_ref_to_Tm_s + Sfus + Cn_l.integrate_by_T_over_T(Tm, T)
    
@functor(var='S.s')
def Solid_Entropy_Ref_Solid(T, Cn_s, T_ref, S0):
    return S0 + Cn_s.integrate_by_T_over_T(T_ref, T)

@functor(var='S.s')
def Solid_Entropy_Ref_Liquid(T, Cn_s, S_int_Tm_to_T_ref_l, Sfus, Tm, S0):
    return S0 - S_int_Tm_to_T_ref_l - Sfus + Cn_s.integrate_by_T_over_T(Tm, T)

@functor(var='S.s')
def Solid_Entropy_Ref_Gas(T, Cn_s, S_int_Tb_to_T_ref_g, Svap_Tb, 
                          S_int_Tm_to_Tb_l, Sfus, Tm, S0):
    return S0 - S_int_Tb_to_T_ref_g - Svap_Tb - S_int_Tm_to_Tb_l - Sfus + Cn_s.integrate_by_T_over_T(Tm, T)
    
@functor(var='S.g')
def Gas_Entropy_Ref_Gas(T, P, Cn_g, T_ref, P_ref, S0):
    return S0 + Cn_g.integrate_by_T_over_T(T_ref, T) - R*log(P/P_ref)

@functor(var='S.g')
def Gas_Entropy_Ref_Liquid(T, P, Cn_g, S_int_T_ref_to_Tb_l, Svap_Tb,
                           Tb, P_ref, S0):
    return S0 + S_int_T_ref_to_Tb_l + Svap_Tb + Cn_g.integrate_by_T_over_T(Tb, T) - R*log(P/P_ref)

@functor(var='S.g')
def Gas_Entropy_Ref_Solid(T, P, Cn_g, S_int_T_ref_to_Tm_s, Sfus,
                          S_int_Tm_to_Tb_l, Svap_Tb, Tb, P_ref, S0):
    return S0 + S_int_T_ref_to_Tm_s + Sfus + S_int_Tm_to_Tb_l + Svap_Tb + Cn_g.integrate_by_T_over_T(Tb, T) - R*log(P/P_ref)


EntropyRefLiquid = PhaseTPFunctorBuilder('S',
                                        Solid_Entropy_Ref_Liquid.functor,
                                        Liquid_Entropy_Ref_Liquid.functor,
                                        Gas_Entropy_Ref_Liquid.functor)

EntropyRefSolid = PhaseTPFunctorBuilder('S',
                                       Solid_Entropy_Ref_Solid.functor,
                                       Liquid_Entropy_Ref_Solid.functor,
                                       Gas_Entropy_Ref_Solid.functor)

EntropyRefGas = PhaseTPFunctorBuilder('S',
                                     Solid_Entropy_Ref_Gas.functor,
                                     Liquid_Entropy_Ref_Gas.functor,
                                     Gas_Entropy_Ref_Gas.functor)

@functor(var='H.l')
def Excess_Liquid_Enthalpy_Ref_Liquid(T, P):
    return 0

@functor(var='H.l')
def Excess_Liquid_Enthalpy_Ref_Gas(T, P, eos, H_dep_Tb_Pb_g,
                                   H_dep_Tb_P_ref_g, eos_1atm):
    return (H_dep_Tb_Pb_g - H_dep_Tb_P_ref_g
            + eos.to_TP(T, P).H_dep_l - eos_1atm.H_dep_l)
    
@functor(var='H.l')
def Excess_Liquid_Enthalpy_Ref_Solid(T, P):
    return 0
    
@functor(var='H.s')
def Excess_Solid_Enthalpy_Ref_Solid(T, P):
    return 0

@functor(var='H.s')
def Excess_Solid_Enthalpy_Ref_Liquid(T, P):
    return 0

@functor(var='H.s')
def Excess_Solid_Enthalpy_Ref_Gas(T, P):
    return 0
    
@functor(var='H.g')
def Excess_Gas_Enthalpy_Ref_Gas(T, P, eos, H_dep_ref_g):
    return eos.to_TP(T, P).H_dep_g - H_dep_ref_g

@functor(var='H.g')
def Excess_Gas_Enthalpy_Ref_Liquid(T, P, eos, H_dep_T_ref_Pb,
                                   H_dep_ref_l, H_dep_Tb_Pb_g):
    return H_dep_T_ref_Pb - H_dep_ref_l + eos.to_TP(T, P).H_dep_g - H_dep_Tb_Pb_g

@functor(var='H.g')
def Excess_Gas_Enthalpy_Ref_Solid(T):
    return 0

ExcessEnthalpyRefLiquid = PhaseTPFunctorBuilder('H_excess',
                                               Excess_Solid_Enthalpy_Ref_Liquid.functor,
                                               Excess_Liquid_Enthalpy_Ref_Liquid.functor,
                                               Excess_Gas_Enthalpy_Ref_Liquid.functor)

ExcessEnthalpyRefSolid = PhaseTPFunctorBuilder('H_excess',
                                                Excess_Solid_Enthalpy_Ref_Solid.functor,
                                                Excess_Liquid_Enthalpy_Ref_Solid.functor,
                                                Excess_Gas_Enthalpy_Ref_Solid.functor)

ExcessEnthalpyRefGas = PhaseTPFunctorBuilder('H_excess',
                                            Excess_Solid_Enthalpy_Ref_Gas.functor,
                                            Excess_Liquid_Enthalpy_Ref_Gas.functor,
                                            Excess_Gas_Enthalpy_Ref_Gas.functor)

@functor(var='S.l')
def Excess_Liquid_Entropy_Ref_Liquid(T, P):
    return 0

@functor(var='S.l')
def Excess_Liquid_Entropy_Ref_Gas(T, P, eos, S_dep_Tb_Pb_g,
                                  S_dep_Tb_P_ref_g, eos_1atm):
    return (S_dep_Tb_Pb_g - S_dep_Tb_P_ref_g
            + eos.to_TP(T, P).S_dep_l - eos_1atm.S_dep_l)
    
@functor(var='S.l')
def Excess_Liquid_Entropy_Ref_Solid(T, P):
    return 0
    
@functor(var='S.s')
def Excess_Solid_Entropy_Ref_Solid(T, P):
    return 0

@functor(var='S.s')
def Excess_Solid_Entropy_Ref_Liquid(T, P):
    return 0

@functor(var='S.s')
def Excess_Solid_Entropy_Ref_Gas(T, P):
    return 0
    
@functor(var='S.g')
def Excess_Gas_Entropy_Ref_Gas(T, P, eos, S_dep_ref_g):
    return eos.to_TP(T, P).S_dep_g - S_dep_ref_g

@functor(var='S.g')
def Excess_Gas_Entropy_Ref_Liquid(T, P, eos, S_dep_T_ref_Pb, 
                                  S_dep_ref_l, S_dep_Tb_Pb_g):
    return S_dep_T_ref_Pb - S_dep_ref_l + eos.to_TP(T, P).S_dep_g - S_dep_Tb_Pb_g

@functor(var='S.g')
def Excess_Gas_Entropy_Ref_Solid(T):
    return 0

ExcessEntropyRefLiquid = PhaseTPFunctorBuilder('S_excess',
                                              Excess_Solid_Entropy_Ref_Liquid.functor,
                                              Excess_Liquid_Entropy_Ref_Liquid.functor,
                                              Excess_Gas_Entropy_Ref_Liquid.functor)

ExcessEntropyRefSolid = PhaseTPFunctorBuilder('S_excess',
                                             Excess_Solid_Entropy_Ref_Solid.functor,
                                             Excess_Liquid_Entropy_Ref_Solid.functor,
                                             Excess_Gas_Entropy_Ref_Solid.functor)

ExcessEntropyRefGas = PhaseTPFunctorBuilder('S_excess',
                                           Excess_Solid_Entropy_Ref_Gas.functor,
                                           Excess_Liquid_Entropy_Ref_Gas.functor,
                                           Excess_Gas_Entropy_Ref_Gas.functor)


# -*- coding: utf-8 -*-
"""
Created on Thu Apr 29 08:43:40 2021

@author: yrc2
"""
import flexsolve as flx

__all__ = ('ideal_coefficient',)

def ideal(cls):
    cls.f = ideal_coefficient
    cls.args = ()
    return cls

@property
def ideal_coefficient(self):
    return _ideal_coefficient

@flx.njitable(cache=True)
def _ideal_coefficient(z=None, T=None, P=None):
    return 1.